import json
import sys
import math
import subprocess

from drawtetrado.svg_painter import Point, ConnType, ConnFlow

class Nucleotide:
    def FindConnections(self, used):
        # Find to what it is connected.
        connected_to = ""
        curr_min = sys.maxsize
        for name, nucl in used.items():
            if nucl["index"] > self.index and \
               curr_min > nucl["index"] and \
               self.chain == nucl["chain"]:
                curr_min = nucl["index"]
                connected_to = name
        return connected_to

    def Block(self, width, height, angle):
        self.coords = []
        self.coords.append(Point(0.0, 0.0))
        if math.tan(math.radians(angle)) != 0:
            shift = height / math.tan(math.radians(angle))
        else:
            shift = height / math.tan(math.radians(0.01))
        self.coords.append(Point(shift, -height))
        self.coords.append(Point(shift + width, -height))
        self.coords.append(Point(width, 0.0))
        self.center = Point((shift + width) / 2.0, -height / 2.0)

    def ShiftBlock(self, shift):
        self.coords[0] += shift
        self.coords[1] += shift
        self.coords[2] += shift
        self.coords[3] += shift
        self.center += shift

    def CalculateCoordinates(self, conf):
        shift = Point(0, 0)
        tan_val = math.tan(math.radians(conf.angle))
        sin_val = math.sin(math.radians(conf.angle))
        cos_val = math.cos(math.radians(conf.angle))
        tan_val_rev = math.tan(math.radians(90.0 - conf.angle))

        shift.y -= self.tetrade_no * (sin_val * (conf.longer + conf.shorter + conf.spacing) + \
                                      conf.tetrade_spacing)
        if self.GetOnzPlusMinus() == "-":
            width_0 = conf.longer
            height_0 = sin_val * conf.shorter
            width_1 = conf.shorter
            height_1 = sin_val * conf.longer
        else: # Should be just this: self.onz == "?+": 
            width_0 = conf.shorter
            height_0 = sin_val * conf.longer
            width_1 = conf.longer
            height_1 = sin_val * conf.shorter

        if self.position == 0:
            self.Block(width_0, height_0, conf.angle)
        elif self.position == 1:
            self.Block(width_1, height_1, conf.angle)
            shift.y -= (width_1 + conf.spacing) * sin_val
            shift.x += (width_1 + conf.spacing) * cos_val
        elif self.position == 2:
            self.Block(width_0, height_0, conf.angle)
            shift.y -= (width_0 + conf.spacing) * sin_val
            shift.x += (width_1 + conf.spacing) + \
                       ((width_0 + conf.spacing) * sin_val) * tan_val_rev
        elif self.position == 3:
            self.Block(width_1, height_1, conf.angle)
            shift.x += width_0 + conf.spacing

        self.ShiftBlock(shift)

    def GetOnzPlusMinus(self):
        if self.onz_full[-1] == "-":
            return "-"
        return "+"

    def GetOnz(self):
        res = self.onz_full[0:1].lower() + "_"
        if self.GetOnzPlusMinus() == "-":
            res += "minus"
        else:
            res += "plus"
        # Only o/n/z_plus/minus are allowed
        if res != "o_plus" and res != "o_minus" and \
           res != "n_plus" and res != "n_minus" and \
           res != "z_plus" and res != "z_minus":
            return "onz_default"
        return res

    def __init__(self, data, used_nucl, tetr_no, tetr_onz, pos):
        self.number = data["number"]
        self.short_name = data["shortName"]
        self.full_name = data["fullName"]
        self.chain = data["chain"]
        self.index = data["index"]
        self.onz_full = tetr_onz
        self.onz = self.GetOnz()
        self.bond = data["glycosidicBond"]
        self.tetrade_no = tetr_no
        self.position = pos
        self.connected_to = self.FindConnections(used_nucl)
        self.connected_from = ""
        self.coords = []
        self.center = Point(0, 0)
        self.connection_type = ConnType.UNKNOWN

        # Parameters regarding connection from this nucleotide to another.
        # out - flow starting the connection
        # in - flow ending the conneciton
        self.flow_out = ConnFlow.UNKNOWN
        self.flow_in = ConnFlow.UNKNOWN


        # Priority of drawing.
        self.priority_conn = -1
        self.priority_edge = -1
        self.priority_nucl = -1

    def UpdatePriorities(self, nucleotides):
        # 1
        if self.connection_type == ConnType.LEFT or \
           self.connection_type == ConnType.LEFT_CROSS or \
           self.connection_type == ConnType.FRONT_TO_BACK:
            self.priority_conn = 1
        elif self.connection_type == ConnType.FRONT_BACK_CROSS and \
             (self.position == 1 or self.position == 2):
            self.priority_conn = 1
        elif self.connection_type == ConnType.SAME_LEVEL and \
             (self.flow_out == ConnFlow.DOWN and self.flow_in == ConnFlow.DOWN):
            self.priority_conn = 1
        elif self.connection_type == ConnType.SIMPLE and self.position == 1:
            self.priority_conn = 1
        # 2 + 3
        if self.connection_type == ConnType.SAME_LEVEL and \
             (self.flow_out == ConnFlow.UP and self.flow_in == ConnFlow.DOWN):
            self.priority_nucl = 2
            self.priority_conn = 3
        elif self.connection_type == ConnType.SAME_LEVEL and \
             (self.flow_out == ConnFlow.DOWN and self.flow_in == ConnFlow.UP):
            if self.connected_to != "":
                nucleotides[self.connected_to].priority_nucl = 2
            self.priority_conn = 3
        # 4
        if self.priority_nucl == -1:
            self.priority_nucl = 4
        # 5 (Have to be later to be on top of the nucleotides)
        if self.priority_edge == -1:
            self.priority_edge = 5
        # 6 (all other connections are processed)
        if  self.priority_conn == -1:
            self.priority_conn = 6


class Quadruplex:
    def UsedNucleotides(self, tetrad, nucl):
        used = {}
        for name, data in tetrad.items():
            used[data["nt1"]] = nucl[data["nt1"]]
            used[data["nt2"]] = nucl[data["nt2"]]
            used[data["nt3"]] = nucl[data["nt3"]]
            used[data["nt4"]] = nucl[data["nt4"]]
        return used

    def PrepareNucleotides(self, structure, quadruplex_id, tetrad_id):
        nucl = structure.nucleotides
        if tetrad_id >= 0:
            tetrads = structure.single_tetrads[quadruplex_id][tetrad_id]
        else:
            tetrads = structure.tetrads[quadruplex_id]
        tetrads_order = structure.tetrads_order[quadruplex_id]

        used_nucl = self.UsedNucleotides(tetrads, nucl)

        tetr_no = 0
        for tetrad_name in tetrads_order:
            if tetrad_name in tetrads:
                tetrad = tetrads[tetrad_name]
            else:
                continue
            nt1 = tetrad["nt1"]
            nt2 = tetrad["nt2"]
            nt3 = tetrad["nt3"]
            nt4 = tetrad["nt4"]
            onz = tetrad["onz"]

            self.nucl_quad[nt1] = Nucleotide(nucl[nt1], used_nucl, tetr_no, onz, 0)
            self.nucl_quad[nt2] = Nucleotide(nucl[nt2], used_nucl, tetr_no, onz, 1)
            self.nucl_quad[nt3] = Nucleotide(nucl[nt3], used_nucl, tetr_no, onz, 2)
            self.nucl_quad[nt4] = Nucleotide(nucl[nt4], used_nucl, tetr_no, onz, 3)

            self.tetrads.append([nt1, nt2, nt3, nt4])

            tetr_no = tetr_no + 1


    def GetChainFirstLast(self):
        chains = {}
        for _, nucl in self.nucl_quad.items():
            if nucl.chain not in chains:
                chains[nucl.chain] = {"first": "",\
                                      "last": "", \
                                      "val": sys.maxsize }
        # Analyze 
        curr_name = ""
        curr_min = sys.maxsize
        for name, nucl in self.nucl_quad.items():
            if chains[nucl.chain]["val"] > nucl.index:
                chains[nucl.chain]["val"] = nucl.index
                chains[nucl.chain]["first"] = name

        # Full connected_from variables.
        for chain, data in chains.items():
            curr_name = data["first"]
            next_conn = self.nucl_quad[data["first"]].connected_to
            while next_conn != "":
                self.nucl_quad[next_conn].connected_from = curr_name
                curr_name = next_conn
                next_conn = self.nucl_quad[next_conn].connected_to
            data["last"] = curr_name

        return chains

    def __init__(self, structure, quadruplex_id, tetrad_id = -1):
        self.nucl_quad = {}
        self.tetrads = []
        self.PrepareNucleotides(structure, quadruplex_id, tetrad_id)
        self.chains = self.GetChainFirstLast()
        if tetrad_id >= 0:
            single_tracts = []
            single_tracts.append(structure.tracts[quadruplex_id][tetrad_id])
            self.tracts = single_tracts
        else:
            self.tracts = structure.tracts[quadruplex_id]

    def GetNucleotidesPositions(self):
        lst = [-1] * len(self.nucl_quad)
        for name, nucl in self.nucl_quad.items():
            if nucl.connected_to != "":
              conn_to = self.nucl_quad[nucl.connected_to]
              lst[nucl.tetrade_no * 4 + nucl.position] = conn_to.tetrade_no * 4 + conn_to.position
            else:
              lst[nucl.tetrade_no * 4 + nucl.position] = -1
        return lst

    def GetSameRotations(self):
        lst = [-1] * len(self.tetrads)
        for idx, tetrad in enumerate(self.tetrads):
            for name in tetrad:
                for group, tetrad_tracts in enumerate(self.tracts):
                    for names in tetrad_tracts:
                        for nucl_name in names:
                            if name == nucl_name:
                                lst[idx] = group
        return lst

    def GetAlignments(self):
        lst = [-1] * len(self.nucl_quad)
        for name, nucl in self.nucl_quad.items():
            for group_1, tetrad_tracts in enumerate(self.tracts):
                for group_2, names in enumerate(tetrad_tracts):
                    for nucl_name in names:
                        if name == nucl_name:
                            lst[nucl.tetrade_no * 4 + nucl.position] = group_1 * 4 + group_2
        return lst

    # Use C++ code to rotate tetrads for more readable output.
    def Optimize(self, optimizer = "./svg_optimizer"):
        import optimizer
        optimized = optimizer.solve(self.GetNucleotidesPositions(),
                                    self.GetSameRotations(),
                                    self.GetAlignments())

        # Update position for nucleotide.
        for _, nucl in self.nucl_quad.items():
            pos = nucl.position + nucl.tetrade_no * 4
            for x in range(4):
                if int(optimized[nucl.tetrade_no * 4 + x]) == nucl.position:
                    nucl.position = x
                    break

        # Update positions in tetrades. For Tetrade border
        level = 0
        for tetrad in self.tetrads:
            new_positions = []
            new_positions.append(tetrad[int(optimized[level * 4 + 0])])
            new_positions.append(tetrad[int(optimized[level * 4 + 1])])
            new_positions.append(tetrad[int(optimized[level * 4 + 2])])
            new_positions.append(tetrad[int(optimized[level * 4 + 3])])

            #print("{0} -> {1}".format(tetrad, new_positions))
            for i, val in enumerate(new_positions):
                tetrad[i] = new_positions[i]

            level = level + 1

    def DetermineConnectionTypes(self):
        for _, nucl in self.nucl_quad.items():
            # Check if it is not the last nucleotide.
            if nucl.connected_to != "":
                conn = self.nucl_quad[nucl.connected_to]
                # Tetrades are counted from bottom to top.
                # > 0 - connection from top to bottom
                # < 0 - connection from bottom to top
                # = 0 - COnnection on the same level
                level_difference = conn.tetrade_no - nucl.tetrade_no
                if conn.position == nucl.position and abs(level_difference) == 1:
                    nucl.connection_type = ConnType.SIMPLE
                elif level_difference == 0:
                    nucl.connection_type = ConnType.SAME_LEVEL
                elif (nucl.position == 2 and conn.position == 2) or \
                     (nucl.position == 3 and conn.position == 3):
                    nucl.connection_type = ConnType.RIGHT
                elif (nucl.position == 2 and conn.position == 3) or \
                     (nucl.position == 3 and conn.position == 2):
                    nucl.connection_type = ConnType.RIGHT_CROSS
                elif (nucl.position == 0 and conn.position == 0) or \
                     (nucl.position == 1 and conn.position == 1):
                    nucl.connection_type = ConnType.LEFT
                elif (nucl.position == 0 and conn.position == 1) or \
                     (nucl.position == 1 and conn.position == 0):
                    nucl.connection_type = ConnType.LEFT_CROSS
                elif (nucl.position == 0 and conn.position == 3) or \
                     (nucl.position == 3 and conn.position == 0) or \
                     (nucl.position == 1 and conn.position == 2) or \
                     (nucl.position == 2 and conn.position == 1):
                    nucl.connection_type = ConnType.FRONT_BACK_CROSS
                elif (nucl.position == 1 and conn.position == 3) or \
                     (nucl.position == 3 and conn.position == 1) or \
                     (nucl.position == 0 and conn.position == 2) or \
                     (nucl.position == 2 and conn.position == 0):
                    nucl.connection_type = ConnType.FRONT_TO_BACK


    def PrintFlow(self, chain):
        nucl = self.nucl_quad[chain["first"]]
        while nucl.connected_to != "":
            conn = self.nucl_quad[nucl.connected_to]
            #print(nucl.full_name, nucl.flow_out, nucl.flow_in)
            nucl = conn

    # TODO Optimize. Do I rly need so many iterations?
    # I HAVE TO FIX IT!! it "works"
    def CalculateFlow(self, chain):
        for _, nucl in self.nucl_quad.items():
            # Check if it is not the last nucleotide.
            if nucl.connected_to != "":
                conn = self.nucl_quad[nucl.connected_to]
                # Tetrades are counted from bottom to top.
                # > 0 - connection from top to bottom
                # < 0 - connection from bottom to top
                # = 0 - COnnection on the same level
                level_difference = nucl.tetrade_no - conn.tetrade_no
                if conn.position == nucl.position and abs(level_difference) == 1:
                    if level_difference == -1:
                        # Connection going bot to top
                        nucl.flow_out = ConnFlow.UP
                        nucl.flow_in = ConnFlow.UP
                        conn.flow_out = ConnFlow.UP
                    else:
                        # Connection top to bot
                        nucl.flow_out = ConnFlow.DOWN
                        nucl.flow_in = ConnFlow.DOWN
                        conn.flow_out = ConnFlow.DOWN


        for _, nucl in self.nucl_quad.items():
            # Check if it is not the last nucleotide.
            if nucl.connected_to != "" and nucl.flow_in == ConnFlow.UNKNOWN:
                conn = self.nucl_quad[nucl.connected_to]
                if conn.flow_out == ConnFlow.UP:
                    nucl.flow_in = ConnFlow.DOWN
                elif conn.flow_out == ConnFlow.DOWN:
                    nucl.flow_in = ConnFlow.UP

        nucl = self.nucl_quad[chain["first"]]
        while nucl.connected_to != "":
            conn = self.nucl_quad[nucl.connected_to]

            if nucl.flow_in == ConnFlow.UNKNOWN:
                level_difference = nucl.tetrade_no - conn.tetrade_no
                if level_difference < 0:
                    # Connection going bot to top
                    nucl.flow_in = ConnFlow.DOWN
                    conn.flow_out = ConnFlow.UP
                elif level_difference > 0:
                    # Connection top to bot
                    nucl.flow_in = ConnFlow.UP
                    conn.flow_out = ConnFlow.DOWN
            nucl = conn

        for _, nucl in self.nucl_quad.items():
            # Check if it is not the last nucleotide.
            if nucl.connected_to != "" and nucl.flow_in == ConnFlow.UNKNOWN:
                conn = self.nucl_quad[nucl.connected_to]
                if conn.flow_out == ConnFlow.UP:
                    nucl.flow_in = ConnFlow.DOWN
                elif conn.flow_out == ConnFlow.DOWN:
                    nucl.flow_in = ConnFlow.UP

        nucl = self.nucl_quad[chain["first"]]
        while nucl.connected_to != "":
            conn = self.nucl_quad[nucl.connected_to]
            if nucl.flow_in == ConnFlow.UNKNOWN:
                nucl.flow_in = nucl.flow_out
                if nucl.flow_out == ConnFlow.UP:
                    conn.flow_out = ConnFlow.DOWN
                elif nucl.flow_out == ConnFlow.DOWN:
                    conn.flow_out = ConnFlow.UP
            nucl = conn

        nucl = self.nucl_quad[chain["last"]]
        while nucl.connected_from != "":
            conn = self.nucl_quad[nucl.connected_from]
            if conn.flow_out == ConnFlow.UNKNOWN:
                level_difference = nucl.tetrade_no - conn.tetrade_no
                if level_difference < 0:
                    # Connection going bot to top
                    conn.flow_out = ConnFlow.DOWN
                elif level_difference > 0:
                    # Connection top to bot
                    conn.flow_out = ConnFlow.UP
                else:
                    conn.flow_out = conn.flow_in
            nucl = conn

        for _, nucl in self.nucl_quad.items():
            # Check if it is not the last nucleotide.
            if nucl.connected_to != "" and nucl.flow_in == ConnFlow.UNKNOWN:
                conn = self.nucl_quad[nucl.connected_to]
                if conn.flow_out == ConnFlow.UP:
                    nucl.flow_in = ConnFlow.DOWN
                elif conn.flow_out == ConnFlow.DOWN:
                    nucl.flow_in = ConnFlow.UP

        # Determine starting flow.
        nucl = self.nucl_quad[chain["first"]]
        if nucl.connected_to != "":
            conn = self.nucl_quad[nucl.connected_to]
            level_difference = nucl.tetrade_no - conn.tetrade_no
            if level_difference < 0:
                # Connection going bot to top
                nucl.flow_out = ConnFlow.UP
            elif level_difference > 0:
                # Connection top to bot
                nucl.flow_out = ConnFlow.DOWN
            else:
                nucl.flow_out = nucl.flow_in

            for _, nucl in self.nucl_quad.items():
                # Check if it is not the last nucleotide.
                if nucl.connected_to != "" and nucl.flow_in == ConnFlow.UNKNOWN:
                    conn = self.nucl_quad[nucl.connected_to]
                    if conn.flow_out == ConnFlow.UP:
                        nucl.flow_in = ConnFlow.DOWN
                    elif conn.flow_out == ConnFlow.DOWN:
                        nucl.flow_in = ConnFlow.UP

        nucl = self.nucl_quad[chain["first"]]
        while nucl.connected_to != "":
            conn = self.nucl_quad[nucl.connected_to]
            if nucl.flow_in == ConnFlow.UNKNOWN:
                if nucl.flow_out == ConnFlow.UNKNOWN:
                    if nucl.connected_from == "":
                        nucl.flow_out = ConnFlow.UP
                    else:
                        conn_from = self.nucl_quad[nucl.connected_from]
                        if conn_from.flow_in == ConnFlow.DOWN:
                            nucl.flow_out = ConnFlow.UP
                        else:
                            nucl.flow_out = ConnFlow.DOWN
                nucl.flow_in = nucl.flow_out
                if nucl.flow_out == ConnFlow.UP:
                    conn.flow_out = ConnFlow.DOWN
                elif nucl.flow_out == ConnFlow.DOWN:
                    conn.flow_out = ConnFlow.UP
            nucl = conn

        #print("\n\n")
        #self.PrintFlow(chain)

class Structure:
    def __init__(self):
        self.nucleotides = {}
        # Index represents corepsponding quadruplex ID data.
        self.tetrads = []
        self.tetrads_order = []
        self.single_tetrads = []
        self.tracts = []



    """
    data - json dict of tetrade parameters.
    {
      "id":"A.DC2-A.DG10-B.DC9-B.DG3",
      "nt1":"A.DC2",
      "nt2":"A.DG10",
      "nt3":"B.DC9",
      "nt4":"B.DG3",
      "onz":"N+",
      "gbaClassification":"VIIIa",
      "planarityDeviation":0.1270031987786134,
      "ionsChannel":[
      ],
      "ionsOutside":[
      ]
    },
    """
    #def addTetrade(self, name, data, tetrade_no):
    #    self.tetrads[tetrade_no][name] = data

    """
    data - json dict of nucleotide parameters.
    {
      "index":22,
      "model":1,
      "chain":"B",
      "number":11,
      "icode":" ",
      "molecule":"DNA",
      "fullName":"B.DG11",
      "shortName":"G",
      "chi":-89.97,
      "glycosidicBond":"syn"
    }
    """
    def addNucleotide(self, name, data):
        self.nucleotides[name] = data

    def fromFile(self, path):
        with open(path) as file:
            data = json.load(file)
        return self.fromJsonDict(data)

    def fromString(self, json_string):
        data = json.loads(json_string)
        return self.fromJsonDict(data)

    def fromJsonDict(self, json_dict):
        for data in json_dict["nucleotides"]:
            self.addNucleotide(data["fullName"], data)


        # TODO What to do with other helices/quadruplexes?
        for index, helice in enumerate(json_dict["helices"]): #["quadruplexes"][0]["tetrads"].items():
            # Unordered tetrads. Order them from "tetrad_pairs"
            single_tetrads_local = []
            tetrad_unordered = {}
            tracts_all = []
            for _, quadruplex in enumerate(helice["quadruplexes"]):
                tetrad_unordered_local = {}
                for data in quadruplex["tetrads"]:
                    tetrad_unordered[data["id"]] = data
                    tetrad_unordered_local[data["id"]] = data
                single_tetrads_local.append(tetrad_unordered_local)
                if "tracts" in quadruplex:
                    tracts_all.append(quadruplex["tracts"])
                else:
                    tracts_all.append(list())

            # Order tetrads according to "tetrad_pairs" data.
            tetrad_pairs = helice["tetradPairs"]
            tetrad_ordered = []
            if tetrad_pairs != None and len(tetrad_pairs) > 0:
                # Add first pair.
                pair = tetrad_pairs.pop(0)
                tetrad_ordered.append(pair["tetrad1"])
                tetrad_ordered.append(pair["tetrad2"])
                to_remove = 0
                while len(tetrad_pairs) > 0:
                    found = False
                    for index_pair, pair in enumerate(tetrad_pairs):
                        for index, tetrad in enumerate(tetrad_ordered):
                            if tetrad == pair["tetrad1"]:
                                tetrad_ordered.insert(index + 1, pair["tetrad2"])
                                found = True
                                break
                            elif tetrad == pair["tetrad2"]:
                                tetrad_ordered.insert(index, pair["tetrad1"])
                                found = True
                                break
                        if found == True:
                            to_remove = index_pair
                            break
                    tetrad_pairs.pop(to_remove)
                    if len(tetrad_pairs) == 0:
                        break

            tetrad_ordered.reverse()

            # Do not add single tetrads as quadruplexes.
            if len(tetrad_ordered) > 1:
                self.tetrads_order.append(tetrad_ordered)
                self.tetrads.append(tetrad_unordered)

            self.tracts.append(tracts_all)
            self.single_tetrads.append(single_tetrads_local)

        # TODO check if input data was valid.
        # Are there any nucleotides and tetrads?
        return self



import os
import sys
import orjson
import gzip
import json
import tempfile
from typing import TextIO
from argparse import ArgumentParser

from eltetrado.analysis import eltetrado
from eltetrado.model import generate_dto
from eltetrado.structure import read_3d_structure, read_2d_structure

import drawtetrado.structure as structure
import drawtetrado.svg_painter as svg_painter

def Draw(struct, output_file, config = svg_painter.Config(1.0)):
    if len(struct.tetrads) == 0:
        print("No tetrads available in the processed structure!")
    for idx in range(len(struct.tetrads)):
        quadruplex = structure.Quadruplex(struct, idx)
        svg_maker = svg_painter.SvgMaker(config, output_file + "_" + str(idx) + ".svg", \
                quadruplex)

        # OPTIMIZE, Takes argument "optimizer" with location to the optimizer
        # binary. Default is "./svg_optimizer"
        quadruplex.Optimize()

        # Prepare + Draw
        svg_maker.DrawAll()

        # Save
        print("Helix " + str(idx) + " full: " + output_file + "_" + str(idx) + ".svg")
        svg_maker.svg.save(pretty=True)

        # Also draw single tetrads from quadruplex structure.
        if (len(struct.single_tetrads[idx]) > 1):
            for tetrad_idx in range(len(struct.single_tetrads[idx])):
                quadruplex = structure.Quadruplex(struct, idx, tetrad_idx)
                svg_maker = svg_painter.SvgMaker(config, output_file + "_" + str(idx) + \
                        "_" + str(tetrad_idx) + ".svg", quadruplex)

                # OPTIMIZE, Takes argument "optimizer" with location to the optimizer
                # binary. Default is "./svg_optimizer"
                quadruplex.Optimize()

                # Prepare + Draw
                svg_maker.DrawAll()

                # Save
                print("Helix " + str(idx) + ", Quadruplex " + str(tetrad_idx) + ": " + output_file + "_" + str(idx) + ".svg")
                svg_maker.svg.save(pretty=True)

def DrawFromString(json, output_file, config = svg_painter.Config(1.0)):
    Draw(structure.Structure().fromString(json), output_file, config)


def DrawFromFile(filename_json, output_file, config = svg_painter.Config(1.0)):
    Draw(structure.Structure().fromFile(filename_json), output_file, config)

# For ElTetrado
def handle_input_file(path) -> TextIO:
    root, ext = os.path.splitext(path)

    if ext == '.gz':
        root, ext = os.path.splitext(root)
        file = tempfile.NamedTemporaryFile('w+', suffix=ext)
        with gzip.open(path, 'rt') as f:
            file.write(f.read())
            file.seek(0)
    else:
        file = tempfile.NamedTemporaryFile('w+', suffix=ext)
        with open(path) as f:
            file.write(f.read())
            file.seek(0)
    return file

def IsFileJson(path):
    with open(path) as file:
        try:
            data = json.load(file)
            return True
        except:
            print("Provided file is not a .json. Processing file using ElTetrado.")
            return False
    return False

def main():
    parser = ArgumentParser('drawtetrado',
        epilog='The output path is a template. Program will generate drawings for each '
        'individual quadruplex in each helix as well as each full helix.'
        'If we have file with 2 helices with 2 quadruplexes each, if no output_template value '
        'is provided, path and basename of the input file will be used. If it is provided, output_template=/tmp/out, '
        'the resulting files will be /tmp/out_0.svg, with full helix 0, /tmp/out_0_0.svg '
        'and /tmp/out_0_1.svg for each quadruplex in helice 0. Similar files will be created '
        'for helix 1 with /tmp/out_1.svg and /tmp/out_1_0.svg, /tmp/out_1_1.svg.')
    parser.add_argument('-i', '--input', help='path to input PDB, PDBx/mmCIF file or JSON generated by ElTetrado. '
            'If PDB or PDBx/mmCIF file is provided, it will be first analyzed using ElTetrado.')
    parser.add_argument('-o', '--output-template', help='(optional) path to output SVG file template '
            '[default=input file path and basename]')
    parser.add_argument('--config',
            help='(optional) JSON config file containing all parameter changes and individual nucleotide '
            'coloring overrides',
            default=None)
    # ElTetrado options.
    parser.add_argument('-m', '--model', help='(optional, ElTetrado) model number to process', default=1, type=int)
    parser.add_argument('--stacking-mismatch',
                        help='(optional, ElTetrado) a perfect tetrad stacking covers 4 nucleotides; this option can be used with value 1 or '
                             '2 to allow this number of nucleotides to be non-stacked with otherwise well aligned '
                             'tetrad [default=2]',
                        default=2,
                        type=int)
    parser.add_argument('--strict',
                        action='store_true',
                        help='(optional, ElTetrado) nucleotides in tetrad are found when linked only by cWH pairing')
    parser.add_argument('--no-reorder',
                        action='store_true',
                        help='(optional, ElTetrado) chains of bi- and tetramolecular quadruplexes should be reordered to be able to have '
                             'them classified; when this is set, chains will be processed in original order, which for '
                             'bi-/tetramolecular means that they will likely be misclassified; use with care!')

    args = parser.parse_args()

    config = svg_painter.Config(1.0, args.config)

    if not args.input:
        print(parser.print_help())
        sys.exit(1)

    root, ext = os.path.splitext(args.input)
    if IsFileJson(args.input):
        if  not args.output_template:
            DrawFromFile(args.input, root, config)
        else:
            DrawFromFile(args.input, args.output_template, config)
    else:
        cif_or_pdb = handle_input_file(args.input)
        structure3d = read_3d_structure(cif_or_pdb, args.model)
        structure2d = read_2d_structure(cif_or_pdb, args.model)
        analysis = eltetrado(structure2d, structure3d, args.strict, args.no_reorder, args.stacking_mismatch)
        dto = generate_dto(analysis)
        if  not args.output_template:
            DrawFromString(orjson.dumps(dto), root, config)
        else:
            DrawFromString(orjson.dumps(dto), args.output_template, config)




# cython: language_level=3
from libcpp.pair cimport pair
from libcpp.vector cimport vector

cdef extern from "svg_optimizer.cpp":
    cdef cppclass Solution:
        vector[int] positions
        int score
        vector[pair[int,int]] dangling_edges

    Solution SolveFailsafe(vector[int]& edges, vector[int]& rotations, vector[int]& alignments)



import svgwrite
import sys
import math
from enum import Enum
import cairo
import tempfile
import json

class ConnType(Enum):
    SIMPLE = 1
    SAME_LEVEL = 2
    RIGHT = 4
    RIGHT_CROSS = 5
    LEFT = 6
    LEFT_CROSS = 7
    FRONT_BACK_CROSS = 11
    FRONT_TO_BACK = 12

    UNKNOWN = -1

class ConnFlow(Enum):
    UP = -1
    DOWN = 1
    UNKNOWN = -1.001 # different value but act as UP

class Side(Enum):
    RIGHT = 1.0
    LEFT = -1.0
    NONE = 0.0

class Point(tuple):
    def __new__(self, x = 0, y = 0):
        return tuple.__new__(Point, (x, y))

    def __init__(self, x = 0, y = 0):
        self.x = x
        self.y = y

    def __add__(self, rhs):
        return Point(self.x + rhs.x, self.y + rhs.y)

    def Distance(self, point):
        return math.sqrt(pow(point.y - self.y, 2.0) +
                    pow(point.x - self.x, 2.0))


class Config:
    def __init__(self, scale = 1.0, config_path = None):
        if config_path == None:
            json_data = json.loads("{}");
        else:
            with open(config_path) as file:
                json_data = json.load(file)
        self.scale = scale if not "scale" in json_data else json_data["scale"]
        # Size parameters of the nucleotides and tetrade.
        self.longer = self.scale * (100.0 if not "nucl-longer" in json_data else json_data["nucl-longer"])
        self.shorter = self.scale * (70.0 if not "nucl-shorter" in json_data else json_data["nucl-shorter"])
        self.spacing = self.scale * (10.0 if not "nucl-spacing" in json_data else json_data["nucl-spacing"])
        self.angle = 50.0 if not "angle" in json_data else json_data["angle"]
        self.tetrade_spacing = self.scale * (50.0 if not "tetrad-spacing" in json_data else json_data["tetrad-spacing"])
        self.stroke_width = self.scale * (3.0 if not "line-stroke" in json_data else json_data["line-stroke"])
        self.point_size = self.scale * (6.0 if not "point-size" in json_data else json_data["point-size"])
        self.point_stroke = self.scale * (2.0 if not "point-stroke" in  json_data else json_data["point-stroke"])
        # Start End label (5' and 3')
        self.se_label_spacing = self.scale * (20.0 if not "se-label-spacing" in json_data else json_data["se-label-spacing"])
        self.se_label_font_size = self.scale * (24.0 if not "se-label-font-size" in json_data else json_data["se-label-font-size"])
        self.font_family = "Arial, Helvetica" if not "font-family" in json_data else json_data["font-family"]
        self.label_font_size = self.scale * (20.0 if not "nucl-font-size" in json_data else json_data["font-family"])
        self.tilted_labels = True if not "tilted-labels" in json_data else json_data["tilted-labels"]

        self.label_chain = True if not "label-chain" in json_data else json_data["label-chain"]
        self.label_nucleotide_full = True if not "label-nucl-fullname" in json_data else json_data["label-nucl-fullname"]
        self.label_nucleotide = True if not "label-nucl-name" in json_data else json_data["label-nucl-name"]
        if self.label_nucleotide == False and self.label_nucelotide_full == True:
            self.label_nucleotide_full = False
        self.label_number = True if not "label-number" in json_data else json_data["label-number"]

        self.colors = {}
        json_colors = {} if not "colors" in json_data else json_data["colors"]
        self.colors["connection"] = "#000000FF" if not "connection" in json_colors else json_colors["connection"]
        self.colors["border"] = "#E23D28FF" if not "border" in json_colors else json_colors["border"]
        self.colors["text"] = "#000000FF" if not "text" in json_colors else json_colors["text"]
        self.colors["point"] = "#FFFFFFFF" if not "point" in json_colors else json_colors["point"]
        # Those refer to the text color on the nucleotyde block.
        self.colors["anti"] = "#FFFFFFFF" if not "anti" in json_colors else json_colors["anti"]
        self.colors["syn"] = "#000000FF" if not "syn" in json_colors else json_colors["syn"]
        self.colors["n/a"] = "#606060FF" if not "n/a" in json_colors else json_colors["n/a"]
        # Those refer to the fill color on the nucleotyde block.
        # Default alpha is 85% -> D9 
        self.colors["onz_default"] = "#646464D9" if not "onz_default" in json_colors else json_colors["onz_default"]
        self.colors["o_plus"] = "#1F78B4D9" if not "o_plus" in json_colors else json_colors["o_plus"]
        self.colors["o_minus"] = "#A6CEE3D9" if not "o_minus" in json_colors else json_colors["o_minus"]
        self.colors["n_plus"] = "#33A02CD9" if not "n_plus" in json_colors else json_colors["n_plus"]
        self.colors["n_minus"] = "#B2DF8AD9" if not "n_minus" in json_colors else json_colors["n_minus"]
        self.colors["z_plus"] = "#FF7F00D9" if not "z_plus" in json_colors else json_colors["z_plus"]
        self.colors["z_minus"] = "#FDBF6FD9" if not "z_minus" in json_colors else json_colors["z_minus"]
        # Allow for ONZ colors above to be overriden if necessary for 
        # individual nucleotides
        # "nucl.full_name": "RGBA"
        self.nucl_colors = {} if not "nucl-color-override" in json_data else json_data["nucl-color-override"]

class SvgMaker:
    def PrepareMarker(self):
        arrowhead = self.svg.marker(insert = (7.6, 4.1), size = (8.0, 6.4), \
                orient = "auto", markerUnits = "strokeWidth", id = "arrowhead")
        arrowhead.add(self.svg.polyline([(0.8, 1.6), (7.6, 4.1), (0.8, 6.4)], \
            stroke = "none", fill = self.GetColor("connection"), \
            fill_opacity = self.GetAlpha("connection")))
        self.svg.defs.add(arrowhead)

    def GetCanvasSize(self, padding, height, width):
        return Point(width + padding * 2, padding + height)


    def __init__(self, config, file_path, quadruplex):
        self.config = config
        self.quadruplex = quadruplex
        sin_val = math.sin(math.radians(config.angle))
        cos_val = math.cos(math.radians(config.angle))

        padding = (config.longer + config.shorter + config.spacing) * sin_val
        padding *= max(len(quadruplex.tetrads) / 5.0, 1.0)
        width = (config.longer + config.shorter + config.spacing) * (cos_val + 1)
        height = ((config.longer + config.shorter + config.spacing) * sin_val + \
                   config.tetrade_spacing) * len(quadruplex.tetrads)

        self.cairo_tempfile = tempfile.NamedTemporaryFile(suffix=".svg")
        self.svg = svgwrite.Drawing(file_path, size = self.GetCanvasSize(padding, height, width), \
                profile = "full")
        self.PrepareMarker()
        self.base_shift = Point(padding, padding * 0.5 + height)

    def ShiftCoords(self, coords, shift):
        if type(coords) == list:
            shifted = []
            for coord in coords:
                shifted.append(coord + shift)
            return shifted
        else:
            return coords + shift

    def RGBAtoAlpha(self, rgba, def_alpha = 0.85):
        if rgba != None and len(rgba[7:9]) == 2:
            return float(int(rgba[7:9], 16)) / 255.0
        return def_alpha

    def RGBAtoRGB(self, rgba):
        return rgba[0:7]

    # Only used for ONZ nucleotide fill or override of the color.
    # Default opacity is 0.85
    def GetAlpha(self, name):
        return self.RGBAtoAlpha(self.config.colors[name])

    def GetColor(self, name):
        if name in self.config.colors:
            return self.RGBAtoRGB(self.config.colors[name])
        return self.RGBAtoRGB(self.config.colors["n/a"]) # Default color

    def GetNuclOverride(self, name):
        # We are checking it before running this function but just in case
        # there is some error we want to return default_onz.
        if name in self.config.nucl_colors:
            return self.RGBAtoRGB(self.config.nucl_colors[name]), \
                   self.RGBAtoAlpha(self.config.nucl_colors[name])
        return self.RGBAtoRGB(self.config.colors["onz_default"]), \
               self.RGBAtoAlpha(self.config.colors["onz_default"])

    def Prepare(self):
        for name, nucl in self.quadruplex.nucl_quad.items():
            nucl.CalculateCoordinates(self.config)

        self.quadruplex.DetermineConnectionTypes()
        for _, chain in self.quadruplex.chains.items():
            self.quadruplex.CalculateFlow(chain)


        for name, nucl in self.quadruplex.nucl_quad.items():
            nucl.UpdatePriorities(self.quadruplex.nucl_quad)


    def DrawPriority(self, priority):
        for name, nucl in self.quadruplex.nucl_quad.items():
            if nucl.priority_nucl == priority:
                self.DrawNucleotide(nucl)

            if nucl.priority_edge == priority:
                self.DrawTetradeBorder(nucl, self.quadruplex)

            if nucl.priority_conn == priority:
                self.DrawConnection(nucl, self.quadruplex.nucl_quad)

    def DrawAll(self):
        self.Prepare();
        # Tetrade numbering of the nucleotides.
        #  2  3
        # 1  4 
        # Draw things in the priority order (first to last) 1 - 5
        # 1 - Conn at the back + Conn FRONT_TO_BACK +
        #     Conn on the left + Conn SAME_LEVEL (down)
        # 2 - Nucleotides with Conn SAME_LEVEL (up-down, down-up)
        # 3 - Connection SAME_LEVEL (up-down, down-up).
        # 4 - Nucleotides Rest
        # 5 - Tetrade border
        # 6 - Connection SAME_LEVEL rest + Rest of connections
        self.DrawPriority(1)
        self.DrawPriority(2)
        self.DrawPriority(3)
        self.DrawPriority(4)
        self.DrawPriority(5)
        self.DrawPriority(6)

        # Draw corner points. As last thing to cover connection 
        # (or should it be earlier)
        for name, nucl in self.quadruplex.nucl_quad.items():
            self.DrawNucleotidePoint(nucl)
            self.DrawNucleotideLabel(nucl)

        # Draw first and last label.
        for _, chain in self.quadruplex.chains.items():
            if chain["first"] != chain["last"]:
                self.DrawLabel(self.quadruplex.nucl_quad[chain["first"]], "5'")
                self.DrawLabel(self.quadruplex.nucl_quad[chain["last"]], "3'")

    def DrawNucleotide(self, nucl):
        shift = self.base_shift
        conf = self.config

        color = self.GetColor(nucl.onz)
        alpha = self.GetAlpha(nucl.onz)
        # Override colors of the nucleotide if it exists in the nucl_color dict.
        if nucl.full_name in conf.nucl_colors:
            color, alpha = self.GetNuclOverride(nucl.full_name)

        block = self.svg.polygon(self.ShiftCoords(nucl.coords, shift), \
                fill = color, fill_opacity = alpha, \
                stroke = color, stroke_width = conf.stroke_width)

        self.svg.add(block)

    def NuclFromFullname(self, nucl):
        fullname = nucl.full_name
        start = max(fullname.find(".") + 1, 0)
        end = fullname.rfind(str(nucl.number))
        if end == -1:
            end = len(fullname) - 1

        return fullname[start:end]


    def NucleotideName(self, nucl):
        # <chain>.<nucleotide-full><number>
        name = ""
        conf = self.config
        if conf.label_chain == True:
            name += str(nucl.chain)
            if conf.label_nucleotide == True or conf.label_number == True:
                name += "."
        if conf.label_nucleotide == True:
            if conf.label_nucleotide_full == True:
                name += str(self.NuclFromFullname(nucl))
            else:
                name += str(nucl.short_name)

        if conf.label_number == True:
            name += str(nucl.number)

        return name

    def ProperFontSize(self, text, fontsize, font, desired_width):
        surface = cairo.SVGSurface(self.cairo_tempfile.name, 1280, 200)
        cr = cairo.Context(surface)
        cr.select_font_face(font, cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_BOLD)
        cr.set_font_size(fontsize)
        xbearing, ybearing, width, height, xadvance, yadvance = cr.text_extents(text)
        while width > desired_width:
            fontsize -= 0.25
            cr.set_font_size(fontsize)
            xbearing, ybearing, width, height, xadvance, yadvance = cr.text_extents(text)

        return fontsize

    def DrawNucleotideLabel(self, nucl):
        shift = self.base_shift
        conf = self.config
        name = self.NucleotideName(nucl)
        font_family = conf.font_family
        font_size = self.ProperFontSize(name, conf.label_font_size, font_family, conf.longer)

        nucl.center = self.ShiftCoords(nucl.center, shift)
        tan_val = math.tan(math.radians(conf.angle))
        sin_val = math.sin(math.radians(conf.angle))

        if (nucl.GetOnzPlusMinus() == "+" and (nucl.position == 0 or nucl.position == 2)) or \
           (nucl.GetOnzPlusMinus() == "-" and (nucl.position == 1 or nucl.position == 3)):
            nucl.center.x += font_size * (sin_val - 0.5)

            rotation = int(360 - conf.angle)
            if conf.tilted_labels == True:
                skewX = 90 - conf.angle
            else:
                skewX = 0
        else:
            nucl.center.y += font_size * (sin_val - 0.5)  # Movie it a little down
            if conf.tilted_labels:
                nucl.center.x += font_size * (sin_val - 1)
            rotation = 0
            if conf.tilted_labels == True:
                skewX = 90 + conf.angle
            else:
                skewX = 0

        outer_color = self.GetColor(nucl.onz)
        # Override colors of the nucleotide if it exists in the nucl_color dict.
        if nucl.full_name in conf.nucl_colors:
            outer_color, _ = self.GetNuclOverride(nucl.full_name)

        label_outline = self.svg.text(name, fill = outer_color, \
                transform = "translate({0}, {1}) rotate({2}) skewX({3})".format( \
                nucl.center.x, nucl.center.y, rotation, skewX, rotation), \
                style = "text-anchor:middle", \
                font_weight = "bold", font_size = font_size, font_family = font_family, \
                stroke = outer_color, stroke_width = "2px", \
                stroke_linejoin = "round")

        label_fill = self.svg.text(name, fill = self.GetColor(nucl.bond), \
                transform = "translate({0}, {1}) rotate({2}) skewX({3})".format( \
                nucl.center.x, nucl.center.y, rotation, skewX), \
                style = "text-anchor:middle", \
                font_weight = "bold", font_size = font_size, font_family = font_family)

        self.svg.add(label_outline)
        self.svg.add(label_fill)


    def DrawTetradeBorder(self, nucl_a, quad):
        point_a = self.ShiftCoords(nucl_a.coords[nucl_a.position], self.base_shift)
        nucl_b = quad.nucl_quad[quad.tetrads[nucl_a.tetrade_no][(nucl_a.position + 1) % 4]]
        point_b = self.ShiftCoords(nucl_b.coords[nucl_b.position], self.base_shift)


        line = self.svg.polyline([point_a, point_b], stroke = self.GetColor("border"), \
                stroke_width = self.config.stroke_width, fill = "none", \
                stroke_opacity = self.GetAlpha("border"))

        self.svg.add(line)


    # For drawinf 5' and 3' labels.
    def DrawLabel(self, nucl, label):
        pos = nucl.position
        pos_str = self.ShiftCoords(nucl.coords[pos], self.base_shift)
        font_size = self.config.se_label_font_size
        pos_str.y += font_size / 5.0
        spacing = self.config.se_label_spacing
        if pos == 0 or pos == 1:
            pos_str.x -= spacing
            anchor = "text-anchor:end"
        else:
            pos_str.x += spacing
            anchor = "text-anchor:begin"
        label_outline = self.svg.text(label, fill = self.GetColor("text"), \
                transform = "translate({0}, {1})".format(pos_str.x, pos_str.y), \
                style = anchor, font_size = font_size, font_weight = "bold", \
                font_family = self.config.font_family,
                stroke = "white", stroke_width = "2px", stroke_linejoin = "round")

        label_fill = self.svg.text(label, fill = self.GetColor("text"), \
                transform = "translate({0}, {1})".format(pos_str.x, pos_str.y), \
                font_family = self.config.font_family,
                style = anchor, font_size = font_size, font_weight = "bold")

        self.svg.add(label_outline)
        self.svg.add(label_fill)

    # Draw simple stright line from A to B.
    def DrawSimpleLine(self, point_a, point_b):
        conf = self.config
        midpoint = (point_a.y - point_b.y) / 2.0
        if point_a.y > point_b.y:
            midpoint += 6.0 * conf.stroke_width
        else:
            midpoint -= 6.0 * conf.stroke_width

        line = self.svg.polyline([point_a, (point_a.x, point_a.y - midpoint), point_b], \
                stroke = self.GetColor("connection"), stroke_width = conf.stroke_width, \
                fill = "none", marker_mid = "url(#arrowhead)", \
                stroke_opacity = self.GetAlpha("connection"))

        self.svg.add(line)

    def DrawSameLevel(self, point_a, point_b, flow_out, flow_in, divisor):
        conf = self.config
        bezier = self.svg.path(d="M", stroke = self.GetColor("connection"), \
                stroke_width = conf.stroke_width, fill = "none", \
                stroke_opacity = self.GetAlpha("connection"))
        distance = point_a.Distance(point_b)

        bezier.push(point_a)
        bezier.push("C")
        bezier.push(point_a + Point(0, flow_out.value * distance / divisor), \
                    point_b + Point(0, flow_in.value * distance / divisor))
        bezier.push(point_b)
        self.svg.add(bezier)


    def DrawSide(self, point_a, point_b, flow_out, flow_in, side, angle, divisor):
        conf = self.config
        bezier = self.svg.path(d="M", stroke = self.GetColor("connection"), \
                stroke_width = conf.stroke_width, fill = "none", \
                stroke_opacity = self.GetAlpha("connection"))
        distance = point_a.Distance(point_b)

        bezier.push(point_a)
        bezier.push("C")

        # If flows are against each other, increase flow (devcrease divisor).
        # Detect this specific cross.
        if point_a.x != point_b.x:
            # Is A above B?
            if point_a.y < point_b.y and flow_out == ConnFlow.UP:
                if flow_in == ConnFlow.DOWN:
                    divisor = divisor / 1.15
                else:
                    divisor = divisor / 1.1
            elif point_a.y > point_b.y and flow_out == ConnFlow.DOWN:
                if flow_in == ConnFlow.DOWN:
                    divisor = divisor / 1.1
                else:
                    divisor = divisor / 1.15

        radians = math.radians(angle) #  math.radians(conf.angle)
        shift_x = distance * math.cos(radians) / divisor
        shift_y = distance * math.sin(radians) / divisor
        if side != Side.NONE:
            bezier.push(point_a + Point(side.value * shift_x, flow_out.value * shift_y))
            bezier.push(point_b + Point(side.value * shift_x, flow_in.value * shift_y))
        else:
            if point_a.x < point_b.x:
                bezier.push(point_a + Point(shift_x, flow_out.value * shift_y))
                bezier.push(point_b + Point(-shift_x, flow_in.value * shift_y))
            else:
                bezier.push(point_a + Point(-shift_x, flow_out.value * shift_y))
                bezier.push(point_b + Point(shift_x, flow_in.value * shift_y))


        bezier.push(point_b)
        self.svg.add(bezier)

    def DrawConnection(self, nucl_a, nucl_quad):
        if nucl_a.connected_to == "":
            return

        point_a = self.ShiftCoords(nucl_a.coords[nucl_a.position], self.base_shift)
        nucl_b = nucl_quad[nucl_a.connected_to]
        point_b = self.ShiftCoords(nucl_b.coords[nucl_b.position], self.base_shift)

        flow_out = nucl_a.flow_out
        flow_in = nucl_a.flow_in
        if nucl_a.connection_type == ConnType.SIMPLE:
            self.DrawSimpleLine(point_a, point_b)
        elif nucl_a.connection_type == ConnType.SAME_LEVEL:
            if (nucl_b.position == 0 and nucl_a.position == 2) or \
                (nucl_b.position == 2 and nucl_a.position == 0):
                self.DrawSameLevel(point_a, point_b, flow_out, flow_in, 3.5)
            else:
                self.DrawSameLevel(point_a, point_b, flow_out, flow_in, 2.5)
        elif nucl_a.connection_type == ConnType.RIGHT:
            self.DrawSide(point_a, point_b, flow_out, flow_in, Side.RIGHT, \
                    self.config.angle, 3.5)
        elif nucl_a.connection_type == ConnType.LEFT:
            self.DrawSide(point_a, point_b, flow_out, flow_in, Side.LEFT, \
                    self.config.angle, 3.5)
        elif nucl_a.connection_type == ConnType.RIGHT_CROSS:
            self.DrawSide(point_a, point_b, flow_out, flow_in, Side.NONE, \
                    80, 2.25)
        elif nucl_a.connection_type == ConnType.LEFT_CROSS:
            self.DrawSide(point_a, point_b, flow_out, flow_in, Side.NONE, \
                    80, 2.25)
        elif nucl_a.connection_type == ConnType.FRONT_BACK_CROSS:
            self.DrawSide(point_a, point_b, flow_out, flow_in, Side.NONE, \
                    80, 2.25)
        elif nucl_a.connection_type == ConnType.FRONT_TO_BACK:
            self.DrawSide(point_a, point_b, flow_out, flow_in, Side.RIGHT, \
                    45, 1.75)
        elif nucl_a.connection_type == ConnType.UNKNOWN:
            print("WARNINMG: Unknown connection type!")
            #line = self.svg.polyline([point_a, point_b], fill = "none", \
            #        stroke = self.GetColor("connection"), stroke_width = self.config.stroke_width, \
            #        marker_mid = "url(#arrowhead)", stroke_opacity = 0.95)

            #self.svg.add(line)

    def DrawNucleotidePoint(self, nucl):
        #print("draw_point")
        conf = self.config
        point = self.ShiftCoords(nucl.coords[nucl.position], self.base_shift)
        self.svg.add(self.svg.circle(point, r = conf.point_size, \
                stroke = self.GetColor("connection"), stroke_width = conf.point_stroke, \
                stroke_opacity = self.GetAlpha("connection"), \
                fill = self.GetColor("point"), fill_opacity = self.GetAlpha("point")))




from setuptools import find_packages, setup

import os

from Cython.Build import cythonize

with open('README.md') as f:
    long_description = f.read()

os.environ['CFLAGS'] = '-std=c++2a'

setup(name = "drawtetrado",
      version = "1.4.1",
      packages = ['drawtetrado'],
      package_dir = {'': 'src'},
      author = "Michal Zurkowski",
      author_email = "michal.zurkowski@cs.put.poznan.pl",
      description = "Draw simplified, layer diagrams of quadruplexes.",
      long_description = long_description,
      long_description_content_type = 'text/markdown',
      url = "https://github.com/michal-zurkowski/drawtetrado",
      project_urls = {
          'Bug Tracker': 'https://github.com/michal-zurkowski/drawtetrado/issues'
      },
      classifiers = [
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3',
          'Topic :: Scientific/Engineering :: Bio-Informatics'
      ],
      ext_modules = cythonize(["cython/optimizer.pyx"]),
      entry_points = {'console_scripts': ['drawtetrado=drawtetrado.main:main']},
      install_requires = [
          'pycairo', 'svgwrite', 'orjson', 'eltetrado'
      ]
)
